@@ -1,10 +1,10 @@
 ;; -*- no-byte-compile: t; lexical-binding: nil -*-
-(define-package "neocaml" "0.1.0"
+(define-package "neocaml" "0.3.0"
   "Major mode for OCaml code."
   '((emacs "29.1"))
-  :url "http://github.com/bbatsov/neocaml"
-  :commit "0918b878eb92c9916ea95c45d22d6fe5781c6d86"
-  :revdesc "0918b878eb92"
+  :url "https://github.com/bbatsov/neocaml"
+  :commit "40c989e597f12661b7f7a078564cc82a9d57f821"
+  :revdesc "40c989e597f1"
   :keywords '("languages" "ocaml" "ml")
   :authors '(("Bozhidar Batsov" . "bozhidar@batsov.dev"))
   :maintainers '(("Bozhidar Batsov" . "bozhidar@batsov.dev")))
